# include "utils.h"
# include "tree.h"

# include <map>
# include <set>
# include <string>
# include <iostream>
# include <vector>
# include <algorithm>
# include <iterator>

using namespace std;

namespace {

template <typename T>
vector<T> singlev(const T &x)
{
  return vector<T>(1, x);
}

template <typename T>
vector<T> reverse(const vector<T> &v)
{
  vector<T> r;
  reverse_copy(v.begin(), v.end(), back_inserter(r));
  return r;
}

/* all functions must ensure that when it returns, the freespace size is at least this */
const int min_freespace_at_return = 10;

template <typename Container>
int sz(const Container &x)
{
  return x.size();
}

string idt_to_suffix(idt x)
{
  if(x.num() > 0)
    return STR(x);
  for(int i = 0; i < ARRAY_SIZE(primitives); i++)
    if(x == primitives[i].val())
      return STR('_' << primitives[i].name);
  throw error("idt_to_suffix: unknown idt");
}

struct label_gen
{
  label_gen() : n(1) { }
  int lbl(idt i)
  {
    pair<map<idt, int>::iterator, bool> p = rel.insert(make_pair(i, n));
    if(p.second)
      n++;
    return p.first->second;
  }
  int anon() { return n++; }
  map<idt, int> get_map() const { return rel; }
private:
  int n;
  map<idt, int> rel;
};

typedef map<idt, int> senv_t;

senv_t get_senv(const prog_c::main_type &main)
{
  senv_t ret;
  int n = 0;
  for(prog_c::main_type::const_iterator it = main.begin(); it != main.end(); ++it)
    ret[it->self] = n++;
  return ret;
}

struct sp_dependent_obj
{
  virtual int write(std::ostream &out, int sp, int indent) const = 0;
  virtual ~sp_dependent_obj(){}
};

typedef counted_ptr<const sp_dependent_obj> sp_dependent;

inline sp_dependent operator+(sp_dependent x, sp_dependent y)
{
  struct cc : sp_dependent_obj
  {
    cc(sp_dependent x, sp_dependent y) : x(x), y(y) {}
    int write(std::ostream &out, int sp, int indent) const
    {
      int sp1 = x->write(out, sp, indent);
      return y->write(out, sp1, indent);
    }
    sp_dependent x, y;
  };
  return sp_dependent(new cc(x, y));
}

sp_dependent raw_string(const string &s)
{
  struct cc : sp_dependent_obj
  {
    cc(const string &s) : s(s) {}
    int write(ostream &out, int sp, int indent) const { out << s; return sp; }
    string s;
  };
  return sp_dependent(new cc(s));
}

# define TXT(chain) raw_string(STR(chain))

sp_dependent label(const string &s)
{
  return TXT(s << ":\n");
}

sp_dependent clabel(int n)
{
  return label(STR("case " << n));
}

string indirect_caller_label(int narg)
{
  if(narg == 1) return "call";
  return STR("call" << narg);
}

string function_label(idt n)
{
  return STR("fun" << idt_to_suffix(n));
}

sp_dependent jlabel(idt n)
{
  return label(function_label(n));
}

sp_dependent label_cj(idt n, label_gen &lg)
{
  return clabel(lg.lbl(n)) + jlabel(n);
}

sp_dependent relative_to_sp(int n)
{
  struct rc : sp_dependent_obj
  {
    rc(int offset) : n(offset) {}
    int write(ostream &out, int sp, int indent) const
    {
      int ofs = n - sp;
      out << -ofs - 1;
      return sp;
    }
    int n;
  };
  return sp_dependent(new rc(n));
}

sp_dependent bpp(int n)
{
  return TXT("sp[") + relative_to_sp(n) + TXT("]");
}

sp_dependent dyn_bpp(sp_dependent n)
{
  return TXT("sp[") + relative_to_sp(0) + TXT(" - ") + n + TXT("]");
}

sp_dependent indented(sp_dependent content)
{
  struct i : sp_dependent_obj
  {
    i(sp_dependent c) : c(c) {}
    int write(ostream &out, int sp, int indent) const
    {
      for(int i = 0; i < indent * 4; i++)
        out <<  " ";
      return c->write(out, sp, indent);
    }
    sp_dependent c;
  };
  return sp_dependent(new i(content + TXT("\n")));
}

sp_dependent indent(sp_dependent content, int level = 1)
{
  struct cc : sp_dependent_obj
  {
    cc(sp_dependent c, int level) : c(c), level(level) {}
    int write(ostream &out, int sp, int indent) const
    {
      return c->write(out, sp, indent+level);
    }
    sp_dependent c;
    int level;
  };
  return sp_dependent(new cc(content, level));
}

sp_dependent work(int n)
{
  return bpp(n + 1);
}

sp_dependent arg(int n, int narg)
{
  return work(narg - n - 1);
}

vector<idt> call_layout(const vector<idt> &params)
{
  return reverse(params);
}

sp_dependent statement(sp_dependent content)
{
  return indented(content + TXT(";"));
}

sp_dependent assign(sp_dependent lhs, sp_dependent rhs)
{
  return statement(lhs + TXT(" = ") + rhs);
}

sp_dependent non_empty(sp_dependent x)
{
  struct o : sp_dependent_obj
  {
    o(sp_dependent x) : x(x) {}
    int write(ostream &out, int sp, int indent) const
    {
      ostringstream ost;
      int sp1 = x->write(ost, sp, indent);
      if(ost.str().empty())
        statement(TXT(""))->write(out, sp, indent);
      else
        out << ost.str();
      return sp1;
    }
    sp_dependent x;
  };
  return sp_dependent(new o(x));
}

sp_dependent set_sp(int n)
{
  struct setter : sp_dependent_obj
  {
    setter(int n) : n(n) {}
    int write(ostream &out, int sp, int indent) const
    {
      if(sp != n)
      {
        int d = n - sp;
        statement(TXT("sp -= " << d))->write(out, sp, indent);
      }
      return n;
    }
    int n;
  };
  return sp_dependent(new setter(n));
}

sp_dependent goto_(const string &label)
{
  return statement(TXT("goto " << label));
}

sp_dependent jump(idt dest)
{
  return goto_(function_label(dest));
}

sp_dependent jump_indir(sp_dependent dest)
{
  return
    assign(TXT("togo"), dest) +
    goto_("jump");
}

sp_dependent ret_novalue()
{
  return
    set_sp(0) +
    jump_indir(bpp(0));
}

sp_dependent ret(sp_dependent retval, sp_dependent cleanup = TXT(""))
{
  return
    assign(TXT("result"), retval) +
    cleanup +
    ret_novalue();
}

sp_dependent static_var(idt n, const senv_t &senv)
{
  if(n == idt_w())
    return TXT("(W)w");
  else if(n == idt_UNIT())
    return TXT(7777);
  return TXT("statics[" << senv.find(n)->second << "]");
}

sp_dependent zeroth_pap(idt n)
{
  return TXT("(W)zero" << idt_to_suffix(n));
}

sp_dependent heap(int offset)
{
  return TXT("hp[" << offset << "]");
}

sp_dependent dyn_heap(sp_dependent offset)
{
  return TXT("hp[") + offset + TXT("]");
}

sp_dependent labelint(int l)
{
  return TXT(l * 2 + 1);
}

sp_dependent operator|(sp_dependent lhs, sp_dependent rhs)
{
  struct cc : sp_dependent_obj
  {
    cc(sp_dependent x, sp_dependent y) : x(x), y(y) {}
    int write(ostream &out, int sp, int indent) const
    {
      int sp1 = x->write(out, sp, indent);
      int sp2 = y->write(out, sp, indent);
      set_sp(sp1)->write(out, sp2, indent);
      return sp1;
    }
    sp_dependent x, y;
  };
  return sp_dependent(new cc(lhs, rhs));
}

sp_dependent block(sp_dependent content)
{
  return
    indented(TXT("{")) +
    indent(content) +
    indented(TXT("}"));
}

sp_dependent if_then_else(sp_dependent cond, sp_dependent br_t, sp_dependent br_f)
{
  return
    indented(TXT("if(") + cond + TXT(")")) +
    block(
      (br_t +
        indent(
          indented(TXT("}")) +
          indented(TXT("else")) +
          indented(TXT("{")), -1)) |
      br_f);
}

sp_dependent deref_ofs(sp_dependent exp, sp_dependent ofs)
{
  return TXT("((W *)") + exp + TXT(")[") + ofs + TXT("]");
}

sp_dependent ensure_freespace(sp_dependent hplim, sp_dependent top)
{
  return indented(TXT("ENSURE_FREESPACE(") + hplim + TXT(", ") + top + TXT(")"));
}

sp_dependent exit_with(const string &status)
{
  return set_sp(0) + statement(TXT("return make_result(sp, hp, " << status << ")"));
}

sp_dependent relative_bound(int n)
{
  return relative_to_sp(n-1);
}

sp_dependent before_allocate(sp_dependent size, int stack_max, int stack_top)
{
  return ensure_freespace(relative_bound(0) + TXT(" - " << stack_max << " - ") + size, relative_bound(stack_top));
}

sp_dependent stack_check(sp_dependent offset, int stack_top)
{
  return ensure_freespace(relative_bound(0) + TXT(" - ") + offset, relative_bound(stack_top));
}

sp_dependent freespace_check_static(int amount, int stack_top)
{
  return amount < stack_top
    ? TXT("") 
    : ensure_freespace(relative_bound(0) + TXT(" - " << amount), relative_bound(stack_top));
}

sp_dependent freespace_check_after_return_static(int amount, int stack_top, int call_offset)
{
  return amount <= min_freespace_at_return + call_offset
    ? TXT("")
    : freespace_check_static(amount, stack_top);
}

sp_dependent incr_hp(sp_dependent amount)
{
  return statement(TXT("hp += ") + amount);
}

sp_dependent loop_general(sp_dependent n, sp_dependent body, bool multiline)
{
  sp_dependent body_part = multiline ? block(body) : indent(body);
  return 
    block(
        statement(TXT("int i")) +
        indented(TXT("for(i = 0; i < ") + n + TXT("; i++)")) +
        body_part);
}

sp_dependent loop_oneline(sp_dependent n, sp_dependent body)
{
  return loop_general(n, body, false);
}

enum var_type
{
  var_global,
  var_local,
  var_static
};

typedef map<idt, var_type> tenv_t;

var_type type(idt i, const tenv_t &te)
{
  tenv_t::const_iterator r = te.find(i);
  if(r == te.end())
    return var_local;
  return r->second;
}

bool local(idt i, const tenv_t &te)
{
  return type(i, te) == var_local;
}

set<idt> all_prims()
{
  set<idt> prims;
  for(int i = 0; i < ARRAY_SIZE(primitives); i++)
    prims.insert(primitives[i].val());
  return prims;
}

struct primitive_usage
{
  primitive_usage() : used_prims(all_prims()), max_unknown_arg(call::max_unknown_call_argnum) {}
  primitive_usage(const set<idt> &p, int m) : used_prims(p), max_unknown_arg(m) {}
  set<idt> used_prims;
  int max_unknown_arg;
};

void gp_on_var(idt x, set<idt> &used, const set<idt> &prims)
{
  if(prims.count(x))
    used.insert(x);
}

void gp_on_apps(const list<call> &apps, set<idt> &used, int &u_max, const set<idt> &prims)
{
  for(list<call>::const_iterator it = apps.begin(); it != apps.end(); ++it)
  {
    gp_on_var(it->fun, used, prims);
    for(int i = 0; i < sz(it->args); i++)
      gp_on_var(it->args[i], used, prims);
    if(it->type == call::unknown)
      u_max = max(u_max, sz(it->args));
  }
}

primitive_usage gather_primitive_usage(const prog_c &pc)
{
  int max = 1;
  set<idt> used;
  set<idt> prims(all_prims());

  gp_on_apps(pc.main, used, max, prims);
  for(prog_c::defs_type::const_iterator it = pc.defs.begin(); it != pc.defs.end(); ++it)
    gp_on_apps((*it)->apps, used, max, prims);
  return primitive_usage(used, max);
}

struct global_env
{
private:
  typedef map<idt, int> aenv_t;
public:
  global_env() {}
  global_env(const tenv_t &t, const senv_t &s, const aenv_t &a, const primitive_usage &p) : te(t), senv(s), aenv(a), p_usage(p) {}
  int arity(idt x) const
  {
    return lookup<int>(aenv, x, "global_env::arity: unregistered idt");
  }
  int static_idx(idt x) const
  {
    return lookup<int>(senv, x, "global_env::static_idx: unregistered idt");
  }
  var_type itype(idt x) const
  {
    return type(x, te);
  }
  bool local(idt x) const
  {
    return itype(x) == var_local;
  }
  bool prim_used(idt x) const
  {
    return p_usage.used_prims.count(x);
  }
  int max_unknown_arg() const
  {
    return p_usage.max_unknown_arg;
  }
  const tenv_t te;
  const senv_t senv;
  const aenv_t aenv;
  const primitive_usage p_usage;
private:
  template <typename T, typename Map>
  static T lookup(const Map &map, idt x, const string &msg)
  {
    typename Map::const_iterator it = map.find(x);
    if(it == map.end())
      throw error(STR(msg << ": " << x));
    return it->second;
  }
};

int lookup_local_index(idt i, const vector<idt> &layout)
{
  vector<idt>::const_iterator it = find(layout.begin(), layout.end(), i);
  if(it == layout.end())
    throw error("lookup_local failed");
  return it - layout.begin();
}

sp_dependent lookup_local(idt i, const vector<idt> &layout)
{
  return work(lookup_local_index(i, layout));
}

sp_dependent lookup_global(idt i, const global_env &ge)
{
  var_type ty = ge.itype(i);
  switch(ty)
  {
  case var_global: return zeroth_pap(i);
  case var_static: return static_var(i, ge.senv);
  }
  throw error("lookup_global: unknown identifier");
}

sp_dependent lookup(idt i, const vector<idt> &layout, const global_env &ge)
{
  if(ge.local(i))
    return lookup_local(i, layout);
  return lookup_global(i, ge);
}

void check_layout(const vector<idt> &layout)
{
  for(int i = 0; i < sz(layout); i++)
    if(layout[i] == idt())
      throw error("hole in layout");
  for(int i = 0; i < sz(layout); i++)
    for(int j = i + 1; j < sz(layout); j++)
      if(layout[i] == layout[j])
        throw error("duplicate in layout");
}

vector<pair<idt, int> > insert_to_layout(vector<idt> &layout, const vector<idt> &args, const vector<idt> &hint, int offset, bool free, const set<idt> &src)
{
  int n = layout.size();
  set<idt> to_insert(src);
  vector<idt> arg_layout = call_layout(args);
  vector<pair<idt, int> > inserts;
  if(!free)
    for(int i = 0; i < sz(args) && i < n - offset; i++)
    {
      if(layout[offset + i] == idt() && to_insert.count(arg_layout[i]))
      {
        inserts.push_back(make_pair(arg_layout[i], offset + i));
        layout[offset + i] = arg_layout[i];
        to_insert.erase(arg_layout[i]);
      }
    }
  for(int i = 0; i < sz(hint) && i < n; i++)
    if(layout[i] == idt() && to_insert.count(hint[i]))
    {
      inserts.push_back(make_pair(hint[i], i));
      layout[i] = hint[i];
      to_insert.erase(hint[i]);
    }
  for(int i = 0; i < n && !to_insert.empty(); i++)
    if(layout[i] == idt())
    {
      inserts.push_back(make_pair(*to_insert.begin(), i));
      layout[i] = *to_insert.begin();
      to_insert.erase(*to_insert.begin());
    }
  return inserts;
}

void augment_layout(vector<idt> &layout, const vector<idt> &args, const vector<idt> &hint, const tenv_t &te, int offset, bool free = false)
{
  set<idt> to_use;
  for(int i = 0; i < sz(args); i++)
    if(local(args[i], te))
      to_use.insert(args[i]);
  for(int i = 0; i < sz(layout); i++)
    if(layout[i] != idt())
      to_use.insert(layout[i]);
  int n = max((int)to_use.size(), sz(layout));
  layout.resize(n);
  for(int i = 0; i < sz(layout); i++)
    to_use.erase(layout[i]);
  insert_to_layout(layout, args, hint, offset, free, to_use);
}

sp_dependent fill_layout(vector<idt> &layout, const vector<idt> &args, const vector<idt> &hint, int offset, bool free)
{
  map<idt, int> to_move;
  set<idt> to_move_set;
  int nhole = count(layout.begin(), layout.end(), idt());
  for(int i = sz(layout) - nhole; i < sz(layout); i++)
    if(layout[i] != idt())
    {
      to_move[layout[i]] = i;
      to_move_set.insert(layout[i]);
    }
  int n = layout.size() - nhole;
  layout.resize(n);
  vector<pair<idt, int> > moves = insert_to_layout(layout, args, hint, offset, free, to_move_set);
  sp_dependent out = TXT("");
  for(int i = 0; i < sz(moves); i++)
    out = out + assign(work(to_move[moves[i].first]), work(moves[i].second));
  check_layout(layout);
  return out;
}

sp_dependent update_layout(vector<idt> &layout, const vector<idt> &args, const vector<idt> &hint, const tenv_t &te, int offset)
{
  augment_layout(layout, args, hint, te, offset);
  return fill_layout(layout, args, hint, offset, false);
}

sp_dependent update_free_layout(vector<idt> &layout, const vector<idt> &args, const vector<idt> &hint, const tenv_t &te)
{
  augment_layout(layout, args, hint, te, 0, true);
  return fill_layout(layout, args, hint, 0, true);
}

int call_offset(const vector<idt> &layout, const vector<idt> &args, const tenv_t &te)
{
  int n = sz(layout);
  while(n > 0 && layout[n-1] == idt())
    n--;
  vector<idt> image = layout;
  update_free_layout(image, args, vector<idt>(), te);
  return min(n, sz(image)) + 1;
}

vector<idt> make_layout(const vector<idt> &args, const vector<idt> &hint, const tenv_t &te)
{
  vector<idt> r;
  augment_layout(r, args, hint, te, 0);
  return r;
}

vector<idt> make_free_layout(const vector<idt> &args, const vector<idt> &hint, const tenv_t &te)
{
  vector<idt> r;
  augment_layout(r, args, hint, te, 0, true);
  return r;
}

sp_dependent shuffle(const vector<pair<int, int> > &mvs)
{
  sp_dependent out = TXT("");
  typedef list<pair<int, int> > moves_type;
  moves_type moves(mvs.begin(), mvs.end());
  while(!moves.empty())
  {
    pair<int, int> initial = moves.front();
    moves.pop_front();
    if(initial.first == initial.second)
      continue;
    vector<pair<int, int> > loop;
    loop.push_back(initial);
    pair<int, int> cur = initial;
    while(cur.second != initial.first)
    {
      for(moves_type::iterator it = moves.begin(); it != moves.end(); ++it)
        if(it->first == cur.second)
        {
          cur = *it;
          moves.erase(it);
          loop.push_back(cur);
          goto found;
        }
      /* not found */
      for(int i = sz(loop) - 1; i >= 0; i--)
        out = out + assign(work(loop[i].second), work(loop[i].first));
      goto end_chain;
found:;
    }
    { /* loop detected */
      sp_dependent assignments = assign(TXT("W first"), work(initial.first));
      for(int i = sz(loop) - 1; i >= 1; i--)
        assignments = assignments + assign(work(loop[i].second), work(loop[i].first));
      assignments = assignments + assign(work(initial.second), TXT("first"));
      out = out + block(assignments);
    }
end_chain:;
  }
  return out;
}

struct dest_comparer
{
  dest_comparer(const vector<idt> &args, const vector<idt> &layout, int offset) : args(args), layout(layout), offset(offset) {}
  bool operator()(int x, int y) const
  {
    return eval(x) < eval(y);
  }
  int eval(int x) const
  {
    return 2 * (x < sz(layout) && layout[x] == args[x - offset]) +
      (x >= sz(layout));
  }
private:
  const vector<idt> &args;
  const vector<idt> &layout;
  const int offset;
};

sp_dependent load_local_part(const vector<idt> &args, int offset, const vector<idt> &layout, const global_env &ge)
{
  typedef map<idt, vector<int> > locals_type;
  locals_type locals;
  for(int i = 0; i < sz(args); i++)
    if(ge.local(args[i]))
      locals[args[i]].push_back(i + offset);
  vector<pair<int, int> > local_moves;
  vector<pair<int, int> > local_copies;
  for(locals_type::const_iterator it = locals.begin(); it != locals.end(); ++it)
  {
    idt to = it->first;
    const vector<int> &dests = it->second;
    vector<int>::const_iterator fst = max_element(dests.begin(), dests.end(), dest_comparer(args, layout, offset));
    local_moves.push_back(make_pair(lookup_local_index(to, layout), *fst));
    for(vector<int>::const_iterator jt = dests.begin(); jt != dests.end(); ++jt)
      if(jt != fst)
        local_copies.push_back(make_pair(*fst, *jt));
  }
  sp_dependent out = shuffle(local_moves);
  for(int i = 0; i < sz(local_copies); i++)
    out = out + assign(work(local_copies[i].second), work(local_copies[i].first));
  return out;
}

sp_dependent load_local_only(const vector<idt> &args, int offset, const vector<idt> &layout)
{
  return load_local_part(args, offset, layout, global_env());
}

sp_dependent load_global_part(const vector<idt> &args, int offset, const global_env &ge, label_gen &lg)
{
  sp_dependent out = TXT("");
  for(int i = 0; i < sz(args); i++)
    if(!local(args[i], ge.te))
    {
      int pos = i + offset;
      out = out + assign(work(pos), lookup_global(args[i], ge));
    }
  return out;
}

sp_dependent load(const vector<idt> &args, int offset, const vector<idt> &layout, const global_env &ge, label_gen &lg)
{
  return load_local_part(args, offset, layout, ge) + load_global_part(args, offset, ge, lg);
}

struct destination_obj
{
  virtual sp_dependent put(sp_dependent val, sp_dependent cleanup) const = 0;
  virtual ~destination_obj() {}
};

typedef counted_ptr<const destination_obj> destination;

destination null_dest()
{
  struct o : destination_obj
  {
    sp_dependent put(sp_dependent val, sp_dependent cleanup) const { return cleanup; }
  };
  return destination(new o);
}

destination assign_to(sp_dependent var)
{
  struct o : destination_obj
  {
    o(sp_dependent var) : var(var) {}
    sp_dependent put(sp_dependent val, sp_dependent cleanup) const
    {
      return assign(var, val) + cleanup;
    }
    sp_dependent var;
  };
  return destination(new o(var));
}

destination return_it()
{
  struct o : destination_obj
  {
    sp_dependent put(sp_dependent val, sp_dependent cleanup) const
    {
      return ret(val, cleanup);
    }
  };
  return destination(new o);
}

destination escape_and(sp_dependent esc, destination d)
{
  struct o : destination_obj
  {
    o(sp_dependent esc, destination d) : esc(esc), d(d) {}
    sp_dependent put(sp_dependent val, sp_dependent cleanup) const
    {
      return esc + d->put(val, cleanup);
    }
    sp_dependent esc;
    destination d;
  };
  return destination(new o(esc, d));
}

sp_dependent invoke_dest(destination d, sp_dependent val, sp_dependent cleanup = TXT(""))
{
  return d->put(val, cleanup);
}

sp_dependent make_pap(idt fun, const vector<idt> &args, const vector<idt> &layout, destination dest, const global_env &ge, label_gen &lg, int stack_max, int stack_top)
{
  sp_dependent alloc_size = TXT(args.size() + 3);
  sp_dependent out = before_allocate(alloc_size, stack_max, stack_top);
  out = out + assign(heap(0), labelint(lg.lbl(fun)));
  int arity = ge.arity(fun);
  out = out + assign(heap(1), TXT(arity - args.size()));
  out = out + assign(heap(2), TXT(arity));
  for(int i = 0; i < sz(args); i++)
    out = out + assign(heap(3+i), lookup(args[i], layout, ge));
  out = out + invoke_dest(dest, TXT("(W)hp"), incr_hp(alloc_size));
  return out;
}

sp_dependent jump_to_pap_general(sp_dependent fun, const vector<idt> &args, const vector<idt> &layout, int stack_top, int offset, int retlabel, const global_env &ge, label_gen &lg)
{
  int narg = args.size();
  sp_dependent set_return_addr = retlabel ? assign(bpp(offset), labelint(retlabel)) : TXT("");
  return 
    block(
        statement(TXT("W pap, arity")) +
        assign(TXT("pap"), fun) +
        assign(TXT("arity"), deref_ofs(TXT("pap"), TXT("2"))) +
        stack_check(TXT("(arity + " << 1 + offset << ")"), stack_top) +
        assign(TXT("pap"), fun) + /* pap may be affected by a GC */
        load(call_layout(args), offset, layout, ge, lg) +
        set_return_addr +
        loop_oneline(TXT("arity - " << narg),
            assign(dyn_bpp(TXT("(arity - i + " << offset << ")")), deref_ofs(TXT("pap"), TXT("i + 3")))) +
        set_sp(offset) +
        jump_indir(deref_ofs(TXT("pap"), TXT("0"))));
}

sp_dependent jump_to_pap(sp_dependent fun, const vector<idt> &args, const vector<idt> &layout, int stack_top, const global_env &ge, label_gen &lg)
{
  return jump_to_pap_general(fun, args, layout, stack_top, 0, 0, ge, lg);
}

sp_dependent call_pap(sp_dependent fun, const vector<idt> &args, const vector<idt> &layout, label_gen &lg, const global_env &ge, destination dest, int stack_top_old, int stack_max, int stack_top_new, int offset)
{
  int cont = lg.anon();
  return
    jump_to_pap_general(fun, args, layout, stack_top_old, offset, cont, ge, lg) +
    clabel(cont) +
    non_empty(invoke_dest(dest, TXT("result")) + freespace_check_after_return_static(stack_max, stack_top_new, offset));
}

sp_dependent augment_pap_raw(sp_dependent fun, const vector<sp_dependent> &args, destination dest, int stack_max, int old_stack_top)
{
  int narg = args.size();
  sp_dependent set_args = TXT("");
  for(int i = 0; i < narg; i++)
    set_args = set_args + assign(dyn_heap(TXT("size + " << 3 + i)), args[i]);
  sp_dependent alloc_size = TXT("(size + " << 3 + narg << ")");
  return
    block(
        statement(TXT("W pap, size")) +
        assign(TXT("pap"), fun) +
        statement(TXT("size = ((W *)pap)[2] - ((W *)pap)[1]")) +
        before_allocate(alloc_size, stack_max, old_stack_top) +
        assign(TXT("pap"), fun) + /* the value of pap may affected by a GC */
        set_args +
        loop_oneline(TXT("size + 3"),
            assign(dyn_heap(TXT("i")), deref_ofs(TXT("pap"), TXT("i")))) +
        assign(heap(1), heap(1) + TXT(" - " << narg)) +
        invoke_dest(dest, TXT("(W)hp"), incr_hp(alloc_size)));
}

sp_dependent augment_pap(idt fun, const vector<idt> &args, const vector<idt> &layout, const global_env &ge, destination dest, int stack_max, int old_stack_top)
{
  vector<sp_dependent> sargs;
  for(int i = 0; i < sz(args); i++)
    sargs.push_back(lookup(args[i], layout, ge));
  return augment_pap_raw(lookup(fun, layout, ge), sargs, dest, stack_max, old_stack_top);
}

sp_dependent jump_to_known(const string &label, const vector<idt> &args, const vector<idt> &layout, const global_env &ge, label_gen &lg, int offset = 0)
{
  return
    load(call_layout(args), offset, layout, ge, lg) +
    set_sp(offset) +
    goto_(label);
}

sp_dependent call_to_known(const string &label, const vector<idt> &args, const vector<idt> &layout, const global_env &ge, label_gen &lg, int offset, destination dest, int stack_max, int stack_top)
{
  int cont = lg.anon();
  return
    load(call_layout(args), offset, layout, ge, lg) +
    assign(bpp(offset), labelint(cont)) +
    set_sp(offset) +
    goto_(label) +
    clabel(cont) +
    non_empty(invoke_dest(dest, TXT("result")) + freespace_check_after_return_static(stack_max, stack_top, offset));
}

sp_dependent unknown_tailcall_onearg(idt fun, idt arg, const vector<idt> &layout, const global_env &ge, label_gen &lg, int stack_max, int stack_top)
{
  return
    if_then_else(
        deref_ofs(lookup(fun, layout, ge), TXT(1)) + TXT(" == 1"),
        jump_to_pap(lookup(fun, layout, ge), singlev(arg), layout, stack_top, ge, lg),
        augment_pap(fun, singlev(arg), layout, ge, return_it(), stack_max, 1 + layout.size()));
}

sp_dependent unknown_call_onearg(idt fun, idt arg, const vector<idt> &layout, int offset, destination dest, const global_env &ge, label_gen &lg, int stack_max, int stack_top)
{
  return
    if_then_else(
        deref_ofs(lookup(fun, layout, ge), TXT(1)) + TXT(" == 1"),
        call_pap(lookup(fun, layout, ge), singlev(arg), layout, lg, ge, dest, layout.size() + 1, stack_max, stack_top, offset),
        augment_pap(fun, singlev(arg), layout, ge, dest, stack_max, layout.size() + 1));
}

pair<vector<idt>, sp_dependent> gen_unknown_tailcall(idt fun, const vector<idt> &args, const vector<idt> &hint, const global_env &ge, label_gen &lg, int &max_stack)
{
  int narg = args.size();
  vector<idt> as;
  as.push_back(fun);
  as.insert(as.end(), args.begin(), args.end());
  vector<idt> layout = make_layout(as, hint, ge.te);
  int stack_top = layout.size() + 1;
  sp_dependent out;
  if(args.size() == 1)
  {
    out = unknown_tailcall_onearg(fun, args[0], layout, ge, lg, max_stack, stack_top);
    max_stack = max(max_stack, max(stack_top, 2));
  }
  else
  {
    out = jump_to_known(indirect_caller_label(narg), as, layout, ge, lg);
    max_stack = max(max_stack, max(stack_top, narg + 2));
  }
  return make_pair(layout, out);
}

sp_dependent gen_unknown_call(idt fun, const vector<idt> &args, destination dest, bool have_dest, const vector<idt> &hint, vector<idt> &layout, const global_env &ge, label_gen &lg, int &max_stack)
{
  int narg = args.size();
  int top = layout.size() + 1;
  vector<idt> as;
  as.push_back(fun);
  as.insert(as.end(), args.begin(), args.end());
  int offset = call_offset(layout, as, ge.te);
  sp_dependent mov = update_layout(layout, as, hint, ge.te, offset);
  sp_dependent ret;
  if(args.size() == 1)
  {
    ret = unknown_call_onearg(fun, args[0], layout, offset, escape_and(mov, dest), ge, lg, max_stack, top);
    max_stack = offset + 2;
  }
  else
  {
    ret = call_to_known(indirect_caller_label(narg), as, layout, ge, lg, offset, escape_and(mov, dest), max_stack, top);
    max_stack = offset + 2 + narg;
  }
  return ret;
}

pair<vector<idt>, sp_dependent> gen_tail_position(const call &c, const vector<idt> &hint, const global_env &ge, label_gen &lg, int &max_stack)
{
  switch(c.type)
  {
  case call::saturated:
  {
    vector<idt> layout = make_layout(c.args, hint, ge.te);
    max_stack = max(max_stack, max((int)layout.size() + 1, (int)c.args.size() + 1));
    sp_dependent out = jump_to_known(function_label(c.fun), c.args, layout, ge, lg);
    return make_pair(layout, out);
  }
  case call::partial:
  {
    vector<idt> layout = make_free_layout(c.args, hint, ge.te);
    int stack_top = layout.size() + 1;
    max_stack = max(max_stack, stack_top);
    sp_dependent out = make_pap(c.fun, c.args, layout, return_it(), ge, lg, max_stack, stack_top);
    return make_pair(layout, out);
  }
  case call::unknown:
    return gen_unknown_tailcall(c.fun, c.args, hint, ge, lg, max_stack);
  }
  throw error("gen_tail_position: unhandled_case");
}

pair<destination, bool> find_dest(vector<idt> &layout, idt destval)
{
  vector<idt>::iterator it = find(layout.begin(), layout.end(), destval);
  if(it != layout.end())
  {
    *it = idt();
    return make_pair(assign_to(work(it - layout.begin())), true);
  }
  return make_pair(null_dest(), false);
}

sp_dependent gen_nontail_position(const call &c, const vector<idt> &hint, vector<idt> &layout, const global_env &ge, label_gen &lg, int &max_stack)
{
  pair<destination, bool> dest_have = find_dest(layout, c.self);
  destination dest = dest_have.first;
  bool have_dest = dest_have.second;
  switch(c.type)
  {
  case call::saturated:
  {
    int top = layout.size() + 1;
    int offset = call_offset(layout, c.args, ge.te);
    sp_dependent mov = update_layout(layout, c.args, hint, ge.te, offset);
    sp_dependent ret = call_to_known(function_label(c.fun), c.args, layout, ge, lg, offset, escape_and(mov, dest), max_stack, top);
    max_stack = (int)layout.size() + 1 + (int)c.args.size() + 1;
    return ret;
  }
  case call::partial:
  if(!have_dest) return TXT("");
  else {
    sp_dependent mov = update_free_layout(layout, c.args, hint, ge.te);
    int stack_top = layout.size() + 1;
    max_stack = max(max_stack, stack_top);
    return make_pap(c.fun, c.args, layout, escape_and(mov, dest), ge, lg, max_stack, stack_top);
  }
  case call::unknown:
    return gen_unknown_call(c.fun, c.args, dest, have_dest, hint, layout, ge, lg, max_stack);
  }
  throw error("gen_nontail_position: unhandled case");
}

sp_dependent gen_def_head(const vector<idt> &initial_layout, const vector<idt> &layout, int max_stack, const global_env &ge)
{
  return
    freespace_check_static(max_stack, initial_layout.size() + 1) +
    load_local_only(layout, 0, initial_layout);
}

sp_dependent gen_def_body(const vector<idt> &args, const list<call> &apps, const global_env &ge, label_gen &lg)
{
  sp_dependent text;
  typedef list<call>::const_reverse_iterator iter;
  if(apps.empty())
    text = freespace_check_static(min_freespace_at_return, args.size() + 1) +
      ret(arg(args.size() - 1, args.size()));
  else
  {
    vector<idt> rparams(call_layout(args));
    int max_stack = min_freespace_at_return;
    pair<vector<idt>, sp_dependent> map_out = gen_tail_position(apps.back(), rparams, ge, lg, max_stack);
    vector<idt> layout = map_out.first;
    sp_dependent out = map_out.second;
    for(iter it = ++apps.rbegin(); it != apps.rend(); ++it)
      out = gen_nontail_position(*it, rparams, layout, ge, lg, max_stack) + out;
    out = gen_def_head(rparams, layout, max_stack, ge) + out;
    text = out;
  }
  return text;
}

void gen_def(ostream &out, const def_c &def, const global_env &ge, label_gen &lg)
{
  sp_dependent text =
    label_cj(def.self, lg) + 
    indented(TXT("FUNCTION_HEAD(" << def.self << ", " << def.args.size() << ")")) +
    gen_def_body(def.args, def.apps, ge, lg);
  text->write(out, 0, 1);
}

tenv_t get_global_set(const prog_c &prog)
{
  tenv_t ret;
  for(prog_c::defs_type::const_iterator it = prog.defs.begin(); it != prog.defs.end(); ++it)
    ret.insert(make_pair((*it)->self, var_global));
  for(prog_c::main_type::const_iterator it = prog.main.begin(); it != prog.main.end(); ++it)
    ret.insert(make_pair(it->self, var_static));
  for(int i = 0; i < ARRAY_SIZE(primitives); i++)
    ret.insert(make_pair(primitives[i].val(),
        primitives[i].callable
          ? var_global
          : var_static));
  return ret;
}

sp_dependent check_char(sp_dependent exp, label_gen &lg)
{
  return deref_ofs(exp, TXT(0)) + TXT(" == ") +  labelint(lg.lbl(idt_CHAR())) + TXT(" && ") + deref_ofs(exp, TXT(1)) + TXT(" == 1");
}

sp_dependent ret_char(sp_dependent code, label_gen &lg, int stack_max, int stack_top)
{
  sp_dependent alloc_size = TXT(4);
  return
    before_allocate(alloc_size, stack_max, stack_top) +
    assign(heap(0), labelint(lg.lbl(idt_CHAR()))) +
    assign(heap(1), TXT(1)) +
    assign(heap(2), TXT(2)) +
    assign(heap(3), code) +
    ret(TXT("(W)hp"), incr_hp(alloc_size));
}
void gen_fun_Out(ostream &out, label_gen &lg)
{
  sp_dependent text =
    label_cj(idt_Out(), lg) +
    freespace_check_static(min_freespace_at_return, 2) +
    block(
        statement(TXT("W c")) +
        assign(TXT("c"), arg(0, 1)) +
        if_then_else(
            check_char(TXT("c"), lg),
            statement(TXT("cxgr_putchar(((W *)c)[3])")) + ret(TXT("c")),
            exit_with("cxgr_error_out_notchar")));
  text->write(out, 0, 1);
}

void gen_fun_In(ostream &out, label_gen &lg)
{
  sp_dependent text =
    label_cj(idt_In(), lg) +
    block(
        statement(TXT("W n")) +
        statement(TXT("n = cxgr_getchar()")) +
        if_then_else(
            TXT("n == CXGR_EOF"),
            ret(arg(0, 1)),
            ret_char(TXT("n"), lg, min_freespace_at_return, 2)));
  text->write(out, 0, 1);
}

void gen_fun_Succ(ostream &out, label_gen &lg)
{
  sp_dependent body =
    statement(TXT("n = ((W *)c)[3] + 2")) +
    statement(TXT("if(n == 513) n = 1")) +
    ret_char(TXT("n"), lg, min_freespace_at_return, 2);
  sp_dependent text =
    label_cj(idt_Succ(), lg) +
    block(
        statement(TXT("W c, n")) +
        assign(TXT("c"), arg(0, 1)) +
        if_then_else(
            check_char(TXT("c"), lg),
            body,
            exit_with("cxgr_error_succ_notchar")));
  text->write(out, 0, 1);
}

void gen_bool_function(ostream &out, label_gen &lg, bool value, idt i)
{
  sp_dependent text =
    clabel(lg.lbl(i)) +
    freespace_check_static(min_freespace_at_return, 3) +
    ret(arg(value ? 0 : 1, 2));
  text->write(out, 0, 1);
}

void gen_fun_CHAR(ostream &out, label_gen &lg)
{
  sp_dependent text =
    clabel(lg.lbl(idt_CHAR())) +
    freespace_check_static(min_freespace_at_return, 3) +
    if_then_else(
        check_char(arg(1, 2), lg) + TXT(" && ") + deref_ofs(arg(1, 2), TXT("3")) + TXT(" == ") + arg(0, 2),
        statement(TXT("result = (W)zero_TRUE")),
        statement(TXT("result = (W)zero_FALSE"))) +
    ret_novalue();
  text->write(out, 0, 1);
}

void gen_fun_REPORT_OUT_OF_STACK(ostream &out, label_gen &lg)
{
  sp_dependent text =
    jlabel(idt_REPORT_OUT_OF_STACK()) +
    exit_with("cxgr_error_stack_oob");
  text->write(out, 0, 1);
}

vector<idt> dummy_params(int n, idt_gen &gen)
{
  vector<idt> r;
  for(int i = 0; i < n; i++)
    r.push_back(gen.gen());
  return r;
}

sp_dependent ic_dispatcher(int n, sp_dependent val)
{
  sp_dependent body = TXT("");
  for(int i = 1; i <= n; i++)
    body = body + indented(TXT("case " << i << ": goto " << indirect_caller_label(n) << "_" << i << ";"));
  return
    indented(TXT("switch(") + val + TXT(")")) +
    indented(TXT("{")) +
    body +
    indented(TXT("default: goto " << indirect_caller_label(n) << "_default;")) +
    indented(TXT("}"));
}

sp_dependent gen_ic_overapp_branch(int present, int need, label_gen &lg)
{
  idt_gen local_g;
  vector<idt> params = dummy_params(present + 1, local_g);
  vector<idt> rparams = call_layout(params);
  idt r = local_g.gen();
  int max_stack = min_freespace_at_return;
  pair<vector<idt>, sp_dependent> layout_second = gen_unknown_tailcall(
      r,
      vector<idt>(params.begin() + 1 + need, params.end()),
      rparams,
      global_env(),
      lg,
      max_stack);
  vector<idt> layout = layout_second.first;
  sp_dependent second = layout_second.second;
  destination dest = find_dest(layout, r).first;
  vector<idt> args(params.begin() + 1, params.begin() + 1 + need);
  vector<idt> as(params.begin(), params.begin() + 1 + need);
  int top = layout.size() + 1;
  int offset = call_offset(layout, as, global_env().te);
  sp_dependent mov = update_layout(layout, as, rparams, global_env().te, offset);
  sp_dependent first = call_pap(
      lookup_local(params[0], layout),
      vector<idt>(params.begin() + 1, params.begin() + 1 + need),
      layout,
      lg,
      global_env(),
      escape_and(mov, dest),
      layout.size() + 1,
      max_stack,
      top,
      offset);
  max_stack = 1 + layout.size();
  return
    label(STR(indirect_caller_label(present) << "_" << need)) +
    gen_def_head(rparams, layout, max_stack, global_env()) +
    first + second;
}

vector<sp_dependent> ic_args(int n)
{
  vector<sp_dependent> args;
  for(int i = 0; i < n; i++)
    args.push_back(arg(i + 1, n + 1));
  return args;
}

sp_dependent gen_ic_justapp_branch(int n, label_gen &lg)
{
  idt_gen localg;
  vector<idt> params = dummy_params(n + 1, localg);
  vector<idt> layout = call_layout(params);
  return 
    label(STR(indirect_caller_label(n) << "_" << n)) +
    jump_to_pap(arg(0, n + 1), vector<idt>(params.begin() + 1, params.end()), layout, n + 2, global_env(), lg);
}

sp_dependent gen_ic_underapp_branch(int n, label_gen &lg)
{
  return 
    label(STR(indirect_caller_label(n) << "_default")) +
    augment_pap_raw(arg(0, n + 1), ic_args(n), return_it(), n + 2, n + 2);
}

void gen_indirect_caller_multi(int n, ostream &out, label_gen &lg)
{
  sp_dependent text = 
    label(indirect_caller_label(n)) +
    ic_dispatcher(n, deref_ofs(arg(0, n+1), TXT(1)));
  for(int i = 1; i < n; i++)
    text = text + gen_ic_overapp_branch(n, i, lg);
  text = text +
    gen_ic_justapp_branch(n, lg) +
    gen_ic_underapp_branch(n, lg);
  text->write(out, 0, 1);
}

void gen_indirect_caller(int n, ostream &out, label_gen &lg)
{
  if(n != 1)
    gen_indirect_caller_multi(n, out, lg);
}

bool need_CHAR(const global_env &ge)
{
  return
    ge.prim_used(idt_In()) ||
    ge.prim_used(idt_Out()) ||
    ge.prim_used(idt_Succ()) ||
    ge.prim_used(idt_w());
}

void gen_builtins(ostream &out, label_gen &lg, const global_env &ge)
{
  if(ge.prim_used(idt_Out()))
    gen_fun_Out(out, lg);
  if(ge.prim_used(idt_In()))
    gen_fun_In(out, lg);
  if(ge.prim_used(idt_Succ()))
    gen_fun_Succ(out, lg);
  if(ge.prim_used(idt_REPORT_OUT_OF_STACK()))
    gen_fun_REPORT_OUT_OF_STACK(out, lg);
  if(need_CHAR(ge))
  {
    gen_bool_function(out, lg, true, idt_TRUE());
    gen_bool_function(out, lg, false, idt_FALSE());
    gen_fun_CHAR(out, lg);
  }
  for(int n = 1; n <= ge.max_unknown_arg(); n++)
    gen_indirect_caller(n, out, lg);
}

void gen_main(ostream &out, const prog_c::main_type &main, const global_env &ge, label_gen &lg)
{
  typedef prog_c::main_type::const_reverse_iterator iter;
  sp_dependent text = TXT("");
  int max_stack = 2;
  for(iter it = main.rbegin(); it != main.rend(); ++it)
  {
    vector<idt> layout = singlev(it->self);
    text =
        gen_nontail_position(*it, vector<idt>(), layout, ge, lg, max_stack) +
        assign(lookup_global(it->self, ge), work(0)) +
        text;
    if(!layout.empty())
      throw error("gen_main: resulting layout not empty");
  }
  text =
    clabel(0) +
    freespace_check_static(max_stack, 1) +
    text +
    exit_with("cxgr_success");
  text->write(out, 0, 1);
}

void gen_code_prolog(ostream &out, const prog_c::main_type &main)
{
  out <<
    "cxgr_result cxgr_code(W *sp, W *hp, W *statics, cxgr_runtime_info *info)\n"
    "{\n"
    "W result;\n"
    "int togo = 0;\n\n"
    "sp[-1] = 9999;\n"
    "jump:\n"
    "switch(togo >> 1)\n"
    "{\n";
}

void gen_code_epilog(ostream &out)
{
  out <<
    "}\n"
    "unhandled_jump(togo, info);\n"
    "return make_result(sp, hp, cxgr_error_internal);\n"
    "}\n";
}

pair<map<idt, int>, global_env> gen_code(ostream &out, const prog_c &prog)
{
  label_gen lg;
  gen_code_prolog(out, prog.main);
  map<idt, int> aenv(arity_env(prog));
  tenv_t te(get_global_set(prog));
  senv_t senv(get_senv(prog.main));
  primitive_usage usage(gather_primitive_usage(prog));
  global_env ge(te, senv, aenv, usage);
  for(prog_c::defs_type::const_iterator it = prog.defs.begin(); it != prog.defs.end(); ++it)
    gen_def(out, **it, ge, lg);
  gen_builtins(out, lg, ge);
  gen_main(out, prog.main, ge, lg);
  gen_code_epilog(out);
  return make_pair(lg.get_map(), ge);
}

int lookup_caselabel(const map<idt, int> &lenv, idt i)
{
  map<idt, int>::const_iterator it = lenv.find(i);
  if(it == lenv.end())
    throw error(STR("lookup_caselabel: not found: " << i));
  return it->second * 2 + 1;
}

void gen_zeroth_pap(ostream &out, idt i, const map<idt, int> &lenv, int arity)
{
  out << "static const W zero" << idt_to_suffix(i) << "[3] = {" << lookup_caselabel(lenv, i)
    << ", " << arity << ", " << arity << "};\n";
}

void gen_zeroth_paps(ostream &out, const prog_c &prog, const map<idt, int> &lenv, const global_env &ge)
{
  for(prog_c::defs_type::const_iterator it = prog.defs.begin(); it != prog.defs.end(); ++it)
    gen_zeroth_pap(out, (*it)->self, lenv, ge.arity((*it)->self));
  if(need_CHAR(ge))
  {
    gen_zeroth_pap(out, idt_TRUE(), lenv, 2);
    gen_zeroth_pap(out, idt_FALSE(), lenv, 2);
  }
  if(ge.prim_used(idt_In()))
    gen_zeroth_pap(out, idt_In(), lenv, 1);
  if(ge.prim_used(idt_Out()))
    gen_zeroth_pap(out, idt_Out(), lenv, 1);
  if(ge.prim_used(idt_Succ()))
    gen_zeroth_pap(out, idt_Succ(), lenv, 1);
}

void gen_w(ostream &out, const map<idt, int> &lenv, const global_env &ge)
{
  if(ge.prim_used(idt_w()))
    out << "static const W w[4] = {" << lookup_caselabel(lenv, idt_CHAR()) << ", 1, 2, 239};\n";
}

void gen_static_count(ostream &out, const prog_c &prog)
{
  out <<
    "int cxgr_static_count(void)\n"
    "{\n"
    "    return " << prog.main.size() << ";\n"
    "}\n";
}

void gen_prolog(ostream &out)
{
  out << 
    "# include \"cxgr_common.h\"\n"
    "# include <stdio.h>\n\n"
    "# ifdef CXGR_EXTRA_INCLUDE\n"
    "#   include CXGR_EXTRA_INCLUDE\n"
    "# endif\n\n"
    "typedef cxgr_word W;\n\n"
    "static cxgr_space invoke_gc(W *sp, W *hp, int hplim, int stacktop, cxgr_runtime_info *info)\n"
    "{\n"
    "    cxgr_space r = cxgr_gc(sp + stacktop, hp, stacktop - hplim, info);\n"
    "    r.sp -= stacktop;\n"
    "    return r;\n"
    "}\n\n"
    "static void unhandled_jump(int togo, cxgr_runtime_info *info)\n"
    "{\n"
    "    char buf[1000];\n"
    "    sprintf(buf, \"internal error: unhandled jump(togo = %d): this error is due to a miscompilation\", togo);\n"
    "    cxgr_message(buf, info);\n"
    "}\n\n"
    "static cxgr_result make_result(W *sp, W *hp, cxgr_status s)\n"
    "{\n"
    "    cxgr_result r;\n"
    "    r.sp = sp;\n"
    "    r.hp = hp;\n"
    "    r.status = s;\n"
    "    return r;\n"
    "}\n\n"
    "# ifndef FUNCTION_HEAD\n"
    "#   define FUNCTION_HEAD(name, arity)\n"
    "# endif\n\n"
    "# define ENSURE_FREESPACE(hplim, stacktop)                                    \\\n"
    "  if(hp > sp + (hplim))                                                       \\\n"
    "  {                                                                           \\\n"
    "    cxgr_space nsp = invoke_gc(sp, hp, (hplim), (stacktop), info);             \\\n"
    "    sp = nsp.sp;                                                              \\\n"
    "    hp = nsp.hp;                                                              \\\n"
    "  }                                                                           \\\n"
    "  /**/\n";
}

} /* nameless namespace */

void generate(ostream &out, const prog_c &prog)
{
  ostringstream body;
  pair<map<idt, int>, global_env> lenv_ge = gen_code(body, prog);
  map<idt, int> &lenv = lenv_ge.first;
  global_env &ge = lenv_ge.second;

  gen_prolog(out);
  out << "\n";
  gen_zeroth_paps(out, prog, lenv, ge);
  out << "\n";
  gen_w(out, lenv, ge);
  out << "\n";
  out << body.str();
  out << "\n";
  gen_static_count(out, prog);
}
