# include "tree.h"
# include "utils.h"

# include <cstdio>
# include <vector>
# include <list>
# include <utility>
# include <stdexcept>
# include <string>
# include <iostream>
# include <map>
# include <set>

using namespace std;

DEF_RUNTIME_ERROR(eof_reached)

namespace {

template <typename Container>
int sz(const Container &x)
{
  return x.size();
}

void error_at(srcloc loc, const string &msg)
{
  cerr << loc << ": error: " << msg << "\n";
}

void warn(srcloc loc, const string &msg)
{
  cerr << loc << ": warning: " << msg << "\n";
}

void handle_empty_source()
{
  cerr << "error: source file empty\n";
  throw abort_compilation();
}

struct location_stream
{
  location_stream(FILE *in) : s(in), loc() {}
  pair<wchar_t, srcloc> get()
  {
    wint_t v = fgetwc(s);
    if(v == WEOF)
      throw eof_reached();
    pair<wchar_t, srcloc> r = make_pair((wchar_t)v, loc);
    if(v == L'\n')
      loc.next_line();
    else
      loc.next_column();
    return r;
  }
private:
  FILE *s;
  srcloc loc;
};

pair<char, srcloc> gr_get_char(location_stream &ps)
{
  while(true) try
  {
    pair<wchar_t, srcloc> p = ps.get();
    srcloc loc = p.second;
    switch(p.first)
    {
    case L'w':
    case L'ｗ':
      return make_pair('w', loc);
    case L'W':
    case L'Ｗ':
      return make_pair('W', loc);
    case L'v':
    case L'ｖ':
      return make_pair('v', loc);
    }
  }catch(eof_reached)
  {
    return make_pair('e', srcloc());
  }
}

template <typename Dest>
void push_apps(list<Dest> &out, Dest (*conv)(const gr_app &), const vector<pair<bool, srcloc> > &line)
{
  int i = 0; 
  while(i < sz(line))
  {
    srcloc loc = line[i].second;
    int f = 0, a = 0;
    for(; i < sz(line) && line[i].first; i++)
      f++;
    if(i == sz(line))
    {
      error_at(line[i-1].second, "application list terminated without an argument");
      throw abort_compilation();
    }
    for(; i < sz(line) && !line[i].first; i++)
      a++;
    out.push_back(conv(gr_app(f, a, loc)));
  }
}

template <typename T>
T identity(const T &t)
{
  return t;
}

auto_ptr<gr_abs> gr_parse_abs(const vector<pair<bool, srcloc> > &line)
{
  int n;
  for(n = 0; n < sz(line) && !line[n].first; n++);
  vector<pair<bool, srcloc> > apps(line.begin() + n, line.end());
  list<gr_app> body;
  push_apps(body, (gr_app (*) (const gr_app &))identity, apps);
  return auto_ptr<gr_abs>(new gr_abs(n, body.begin(), body.end(), line[0].second));
}

toplevel app_to_toplevel(const gr_app &a)
{
  return toplevel(a);
}

void push_line(gr_prog &prog, const vector<pair<bool, srcloc> > &line)
{
  if(!line.empty() && !line[0].first) /* function definition */
  {
    auto_ptr<gr_abs> a = gr_parse_abs(line);
    auto_ptr<const gr_abs> b(a);
    prog.push_back(toplevel(b));
  }
  else
    push_apps(prog, app_to_toplevel, line);
}

idt lookup_idt(const vector<idt> &env, int index, srcloc loc)
{
  if(index > sz(env))
  {
    warn(loc, "stack index out of bounds (will result in a runtime error)");
    return idt_out_of_stack();
  }
  return env[env.size()-index];
}

struct gr_to_simple_converter : toplevel_visitor
{
  gr_to_simple_converter(idt_gen &gen) : gen(gen)
  {
    genv.push_back(idt_In());
    genv.push_back(idt_w());
    genv.push_back(idt_Succ());
    genv.push_back(idt_Out());
  }
  void visit(const gr_app &app)
  {
    idt ai = gen.gen();
    main.push_back(application(
          lookup_idt(genv, app.fun, app.loc),
          lookup_idt(genv, app.arg, app.loc),
          ai));
    genv.push_back(ai);
  }
  void visit(const gr_abs &abs)
  {
    vector<idt> env(genv);
    vector<idt> params;
    for(int i = 0; i < abs.narg; i++)
    {
      idt pi = gen.gen();
      params.push_back(pi);
      env.push_back(pi);
    }
    typedef list<gr_app>::const_iterator iter;
    vector<application> body;
    for(iter it = abs.apps.begin(); it != abs.apps.end(); ++it)
    {
      idt ai = gen.gen();
      body.push_back(application(
            lookup_idt(env, it->fun, it->loc),
            lookup_idt(env, it->arg, it->loc),
            ai));
      env.push_back(ai);
    }
    idt di = gen.gen();
    defs.push_back(counted_ptr<def_a>(
          new def_a(di, params.begin(), params.end(), body.begin(), body.end())));
    genv.push_back(di);
  }
  void terminate()
  {
    idt ai = gen.gen();
    main.push_back(application(genv.back(), genv.back(), ai));
    genv.push_back(ai);
  }
  prog_a result() const
  {
    return prog_a(defs.begin(), defs.end(), main.begin(), main.end());
  }

private:
  idt_gen &gen;
  vector<idt> genv;
  list<counted_ptr<def_a> > defs;
  list<application> main;
};

list<application> oe_apps(const list<application> &apps)
{
  list<application> r;
  for(list<application>::const_iterator it = apps.begin(); it != apps.end(); ++it)
    if(it->fun() != idt_out_of_stack() && it->arg() != idt_out_of_stack())
      r.push_back(*it);
    else
    {
      r.push_back(application(idt_REPORT_OUT_OF_STACK(), idt_UNIT(), it->result()));
      break;
    }
  return r;
}

prog_a::defs_type oe_defs(const prog_a::defs_type &defs)
{
  prog_a::defs_type r;
  for(prog_a::defs_type::const_iterator it = defs.begin(); it != defs.end(); ++it)
  {
    list<application> apps = oe_apps((*it)->apps);
    r.push_back(counted_ptr<def_a>(new def_a((*it)->self, (*it)->args.begin(), (*it)->args.end(), apps.begin(), apps.end())));
  }
  return r;
}

call collect_apps(list<application>::const_iterator &cur, list<application>::const_iterator end, int &pos,
    idt fun, const vector<idt> &initial_args,
    const map<idt, int> &last_occurence, const map<idt, int> &aenv,
    bool can_contract)
{
  map<idt, int>::const_iterator r = aenv.find(fun);
  bool local = r == aenv.end();
  int arity = local ? call::max_unknown_call_argnum : r->second;

  vector<idt> args = initial_args;
  idt last;
  do {
    args.push_back(cur->arg());
    last = cur->result();
    ++cur;
    pos++;
  } while (can_contract && cur != end && cur->fun() == last && sz(args) < arity &&
        cur->arg() != last && last_occurence.find(last)->second == pos);
  call::calltype ct = local ? call::unknown : sz(args) == arity ? call::saturated : call::partial;
  return call(fun, args.begin(), args.end(), last, ct);
}

list<call> a_to_c_apps(const list<application> src, const map<idt, int> &aenv, bool can_contract)
{
  list<call> out;
  typedef list<application>::const_iterator iter;
  map<idt, int> last_occurence;
  map<idt, const call *> partials;
  int n = 0;
  for(iter it = src.begin(); it != src.end(); ++it)
  {
    last_occurence[it->fun()] = last_occurence[it->arg()] = n;
    n++;
  }
  iter it = src.begin();
  n = 0;
  while(it != src.end())
  {
    call c = partials.count(it->fun())
      ? collect_apps(it, src.end(), n, partials[it->fun()]->fun, partials[it->fun()]->args, last_occurence, aenv, can_contract)
      : collect_apps(it, src.end(), n, it->fun(), vector<idt>(), last_occurence, aenv, can_contract);
    out.push_back(c);
    if(c.type == call::partial)
      partials[c.self] = &out.back();
  }
  return out;
}

def_c a_to_c_def(const def_a &src, const map<idt, int> &aenv)
{
  list<call> apps(a_to_c_apps(src.apps, aenv, true));
  return def_c(src.self, src.args.begin(), src.args.end(), apps.begin(), apps.end());
}

} /* nameless namespace */

gr_prog gr_parse(FILE *in)
{
  location_stream ps(in);
  pair<char, srcloc> c;
  vector<pair<bool, srcloc> > line;
  gr_prog prog;

  do {
    c = gr_get_char(ps);
  } while (c.first == 'W' || c.first == 'v');

  if(c.first == 'e')
    handle_empty_source();

  while(true)
  {
    if(c.first == 'e')
    {
      push_line(prog, line);
      break;
    }
    else if(c.first == 'v')
    {
      push_line(prog, line);
      line.clear();
    }
    else
      line.push_back(make_pair(c.first == 'W', c.second));
    c = gr_get_char(ps);
  }

  return prog;
}

prog_a gr_to_simple(const gr_prog &src, idt_gen &gen)
{
  gr_to_simple_converter conv(gen);
  for(gr_prog::const_iterator it = src.begin(); it != src.end(); ++it)
    it->accept(&conv);
  conv.terminate();
  return conv.result();
}

prog_a make_oob_explicit(const prog_a &prog)
{
  prog_a::defs_type defs(oe_defs(prog.defs));
  prog_a::main_type main(oe_apps(prog.main));
  return prog_a(defs.begin(), defs.end(), main.begin(), main.end());
}

prog_c a_to_c(const prog_a &src)
{
  list<counted_ptr<def_c> > defs;
  map<idt, int> aenv(arity_env(src));
  typedef prog_a::defs_type::const_iterator iter;
  for(iter it = src.defs.begin(); it != src.defs.end(); ++it)
    defs.push_back(counted_ptr<def_c>(new def_c(a_to_c_def(**it, aenv))));
  list<call> main(a_to_c_apps(src.main, aenv, false));
  return prog_c(defs.begin(), defs.end(), main.begin(), main.end());
}

