# ifndef TREE_H
# define TREE_H

# include "utils.h"

# include <cstdlib>
# include <vector>
# include <list>
# include <memory>
# include <iostream>
# include <map>

struct idt
{
  explicit idt(int n = 0) : v(n) {}
  int num() const{ return v; }
  bool operator==(const idt &other) const { return v == other.v; }
  bool operator!=(const idt &other) const { return !(*this == other); }
  bool operator<(const idt &other) const { return v < other.v; }
private:
  int v;
};

inline std::ostream &operator<<(std::ostream &ost, const idt &v)
{
  return ost << v.num();
}

struct primitive_info
{
  idt (*val)();
  const char *name;
  bool callable;
  int arity; /* -1 if not callable */
};

# define DEF_CONST_IDT(name, val) inline idt idt_##name() { return idt((val)); }

DEF_CONST_IDT(In, -4)
DEF_CONST_IDT(w, -3)
DEF_CONST_IDT(Succ, -2)
DEF_CONST_IDT(Out, -1)
DEF_CONST_IDT(out_of_stack, -100)
DEF_CONST_IDT(CHAR, -200)
DEF_CONST_IDT(TRUE, -201)
DEF_CONST_IDT(FALSE, -202)
DEF_CONST_IDT(REPORT_OUT_OF_STACK, -203)
DEF_CONST_IDT(UNIT, -204)
inline idt idt_CALL(int narg) { return idt(-300 - narg); }

# undef DEF_CONST_IDT

const primitive_info primitives[] = {
  { idt_In, "In", true, 1 },
  { idt_w, "w", false, -1 },
  { idt_Succ, "Succ", true, 1 },
  { idt_Out, "Out", true, 1 },
  { idt_out_of_stack, "out_of_stack", false, -1 },
  { idt_CHAR, "CHAR", false, -1 },
  { idt_TRUE, "TRUE", false, -1 },
  { idt_FALSE, "FALSE", false, -1 },
  { idt_REPORT_OUT_OF_STACK, "REPORT_OUT_OF_STACK", true, 1 },
  { idt_UNIT, "UNIT", false, -1 },
};

struct idt_gen
{
  idt_gen() : source(1) {}
  idt gen() { return idt(source++); }
private:
  int source;
  idt_gen(const idt_gen &);
  idt_gen &operator=(const idt_gen &);
};

struct application
{
  application(idt fun, idt arg, idt result) : _fun(fun), _arg(arg), _result(result) {}
  idt fun() const { return _fun; }
  idt arg() const { return _arg; }
  idt result() const { return _result; }
private:
  idt _fun;
  idt _arg;
  idt _result;
};

inline std::ostream &operator<<(std::ostream &ost, const application &app)
{
  return ost << app.result() << " = " << app.fun() << " " << app.arg();
}

struct call
{
  enum calltype
  {
    saturated,
    partial,
    unknown
  };
  template <typename In>
  call(idt fun, In f, In l, idt self, calltype tp)
    :fun(fun), args(f, l), self(self), type(tp) {}
  idt fun;
  std::vector<idt> args;
  idt self;
  calltype type;
  enum{ max_unknown_call_argnum = 7 };
  /* invariant: if type == unknown, then args.size() <= max_unknown_call_argnum */
  /* invariant: results of known partial applications never appear in the fun field of subsequent unknown calls */
};

inline std::ostream &operator<<(std::ostream &ost, const call &c)
{
  ost << c.self << " = ";
  if(c.type == call::saturated)
    ost << "(S)";
  else if(c.type == call::partial)
    ost << "(P)";
  ost << c.fun;
  for(int i = 0; i < (int)c.args.size(); i++)
    ost << " " << c.args[i];
  return ost;
}

template <typename App>
struct definition
{
  template <typename In, typename In1>
  definition(idt self, In args_begin, In args_end, In1 apps_begin, In1 apps_end)
    : self(self), args(args_begin, args_end), apps(apps_begin, apps_end) {}
  const idt self;
  const std::vector<idt> args;
  const std::list<App> apps;
};

template <typename App>
struct program
{
  template <typename In, typename In1>
  program(In defs_f, In defs_l, In1 main_f, In1 main_l)
    :defs(defs_f, defs_l), main(main_f, main_l) {}
  typedef std::list<counted_ptr<definition<App> > > defs_type;
  defs_type defs;
  typedef std::list<App> main_type;
  main_type main;
};

typedef definition<application> def_a;
typedef program<application> prog_a;
typedef definition<call> def_c;
typedef program<call> prog_c;

struct srcloc
{
  explicit srcloc(int line = 1, int col = 1) : line(line), col(col) {}
  void next_column(){ col++; }
  void next_line() { line++; col = 1; }
  int line;
  int col;
};

inline std::ostream &operator<<(std::ostream &ost, const srcloc &loc)
{
  return ost << loc.line << ":" << loc.col;
}

struct gr_app
{
  gr_app(int fun, int arg, srcloc loc) : fun(fun), arg(arg), loc(loc) {}
  const int fun;
  const int arg;
  const srcloc loc;
};

struct gr_abs
{
  template <typename In>
  gr_abs(int narg, In apps_begin, In apps_end, srcloc loc)
    : narg(narg), apps(apps_begin, apps_end), loc(loc) {}
  const int narg;
  const std::list<gr_app> apps;
  const srcloc loc;
};

struct toplevel_visitor
{
  virtual void visit(const gr_app &) = 0;
  virtual void visit(const gr_abs &) = 0;
  virtual ~toplevel_visitor() {}
};

struct toplevel_obj
{
  virtual void accept(toplevel_visitor *) const = 0;
  virtual ~toplevel_obj() {}
};

struct app_toplevel : toplevel_obj
{
  explicit app_toplevel(gr_app app) : app(app) {}
  void accept(toplevel_visitor *tv) const { tv->visit(app); }
private:
  const gr_app app;
};

struct abs_toplevel : toplevel_obj
{
  explicit abs_toplevel(std::auto_ptr<const gr_abs> p) : pabs(p) {}
  void accept(toplevel_visitor *tv) const { tv->visit(*pabs); }
private:
  const std::auto_ptr<const gr_abs> pabs;
};

struct toplevel
{
  explicit toplevel(const gr_app &app) : pobj(new app_toplevel(app)) {}
  explicit toplevel(std::auto_ptr<const gr_abs> p) : pobj(new abs_toplevel(p)) {}
  explicit toplevel(counted_ptr<toplevel_obj> p) : pobj(p) {}
  toplevel(const toplevel &other) : pobj(other.pobj) {}
  void accept(toplevel_visitor *tv) const { pobj->accept(tv); }
private:
  counted_ptr<toplevel_obj> pobj;
};

typedef std::list<toplevel> gr_prog;

gr_prog gr_parse(std::FILE *in);
prog_a gr_to_simple(const gr_prog &prog, idt_gen &gen);
prog_a make_oob_explicit(const prog_a &prog);
prog_c a_to_c(const prog_a &simple);

void generate(std::ostream &out, const prog_c &prog);

DEF_RUNTIME_ERROR(abort_compilation)

template <typename App>
std::map<idt, int> arity_env(const program<App> &src)
{
  std::map<idt, int> aenv;
  typedef typename program<App>::defs_type::const_iterator iter;
  for(iter it = src.defs.begin(); it != src.defs.end(); ++it)
    aenv[(*it)->self] = (*it)->args.size();
  for(int i = 0; i < ARRAY_SIZE(primitives); i++)
    if(primitives[i].callable)
      aenv[primitives[i].val()] = primitives[i].arity;
  for(int i = 2; i <= call::max_unknown_call_argnum; i++)
    aenv[idt_CALL(i)] = i + 1;
  return aenv;
}

template <typename App>
void print_definition(const definition<App> &def)
{
  using namespace std;
  cout << def.self << " = \\";
  for(int i = 0; i < (int)def.args.size(); i++)
    cout << def.args[i] << " ";
  cout << "->";
  for(typename list<App>::const_iterator it = def.apps.begin(); it != def.apps.end(); ++it)
    cout << " " << *it << ";";
  cout << "\n";
}

template <typename App>
void print_program(const program<App> &prog)
{
  using namespace std;
  typedef typename program<App>::defs_type::const_iterator iter;
  for(iter it = prog.defs.begin(); it != prog.defs.end(); ++it)
    print_definition(**it);
  for(typename list<App>::const_iterator it = prog.main.begin(); it != prog.main.end(); ++it)
    cout << *it << "\n";
}

# endif
